  ; ---------------------------------------------------------------------
  ; These two macros perform the grunt work involved in measuring the
  ; processor clock cycle count for a block of code. These macros must
  ; be used in pairs, and the block of code must be placed in between
  ; the clockctr_begin and clockctr_end macro calls. The clockctr_end
  ; macro returns the clock cycle count for a single pass through the
  ; block of code, corrected for the test loop overhead, in EAX.
  ;
  ; If your code is using MMX instructions without executing an EMMS
  ; at the end of each MMX instruction sequence, defining the symbol
  ; _EMMS will cause the clockctr_end macro to insert an EMMS in
  ; front of the FP instructions.
  ;
  ; The loopcount parameter should be set to a relatively high value
  ; to produce repeatable results.
  ;
  ; Note that setting the priority parameter to REALTIME_PRIORITY_CLASS
  ; involves some risk, as it will cause your process to preempt *all*
  ; other processes, including critical Windows processes. In most
  ; cases HIGH_PRIORITY_CLASS will produce the same cycle count.
  ; ---------------------------------------------------------------------
    clockctr_begin MACRO loopcount:REQ, priority
        LOCAL label
        IFNDEF __clockctr__stuff__defined__
          __clockctr__stuff__defined__ equ <1>
          .data
          ALIGN 8             ;; Optimal alignment for QWORD
            __clockctr__qword__count__  dq 0
            __clockctr__loop__count__   dd 0
            __clockctr__loop__counter__ dd 0
          .code
        ENDIF
        mov __clockctr__loop__count__, loopcount
        IFNB <priority>
          invoke GetCurrentProcess
          invoke SetPriorityClass, eax, priority
        ENDIF
        xor   eax, eax        ;; Use same CPUID input value for each call
        cpuid                 ;; Flush pipe & wait for pending ops to finish
        rdtsc                 ;; Read Time Stamp Counter

        push  edx             ;; Preserve high-order 32 bits of start count
        push  eax             ;; Preserve low-order 32 bits of start count
        mov   __clockctr__loop__counter__, loopcount
        xor   eax, eax
        cpuid                 ;; Make sure loop setup instructions finish
      ALIGN 16                ;; Optimal loop alignment for P6
      @@:                     ;; Start an empty reference loop
        sub   __clockctr__loop__counter__, 1
        jnz   @B

        xor   eax, eax        
        cpuid                 ;; Make sure loop instructions finish
        rdtsc                 ;; Read end count
        pop   ecx             ;; Recover low-order 32 bits of start count
        sub   eax, ecx        ;; Low-order 32 bits of overhead count in EAX
        pop   ecx             ;; Recover high-order 32 bits of start count
        sbb   edx, ecx        ;; High-order 32 bits of overhead count in EDX
        push  edx             ;; Preserve high-order 32 bits of overhead count
        push  eax             ;; Preserve low-order 32 bits of overhead count        
        xor   eax, eax
        cpuid
        rdtsc
        push  edx             ;; Preserve high-order 32 bits of start count
        push  eax             ;; Preserve low-order 32 bits of start count
        mov   __clockctr__loop__counter__, loopcount
        xor   eax, eax
        cpuid                 ;; Make sure loop setup instructions finish
      ALIGN 16                ;; Optimal loop alignment for P6
      label:                  ;; Start test loop
        __clockctr__loop__label__ equ <label>
    ENDM

    clockctr_end MACRO
        sub   __clockctr__loop__counter__, 1
        jnz   __clockctr__loop__label__
        xor   eax, eax
        cpuid                 ;; Make sure loop instructions finish
        rdtsc                 ;; Read end count
        pop   ecx             ;; Recover low-order 32 bits of start count
        sub   eax, ecx        ;; Low-order 32 bits of test count in EAX
        pop   ecx             ;; Recover high-order 32 bits of start count
        sbb   edx, ecx        ;; High-order 32 bits of test count in EDX
        pop   ecx             ;; Recover low-order 32 bits of overhead count
        sub   eax, ecx        ;; Low-order 32 bits of adjusted count in EAX
        pop   ecx             ;; Recover high-order 32 bits of overhead count
        sbb   edx, ecx        ;; High-order 32 bits of adjusted count in EDX

        mov   DWORD PTR __clockctr__qword__count__, eax
        mov   DWORD PTR __clockctr__qword__count__ + 4, edx

        invoke GetCurrentProcess
        invoke SetPriorityClass, eax, NORMAL_PRIORITY_CLASS

        IFDEF _EMMS
          EMMS
        ENDIF

        finit
        fild  __clockctr__qword__count__
        fild  __clockctr__loop__count__
        fdiv
        fistp __clockctr__qword__count__

        mov   eax, DWORD PTR __clockctr__qword__count__
    ENDM
